from .calendar_round import *
from .long_count import *
from .tzolkin import *
from .haab import *
from .utils import *

__all__ = ["Mayadate"]

class Mayadate:
    """ Umbrella class to handle Maya calendar dates, conversions, and inference

    Attributes:
        long_count (LongCount): The Long Count representation of the date
        calendar_round (CalendarRound): The Calendar Round position of the date
        glyph_g (str): The Glyph G associated with the date

    """

    def __init__(self, long_count=None, calendar_round=None, glyph_g=None):
        """Creates a new Mayadate object

        Args:
            long_count (LongCount): The Long Count representation of the date
            calendar_round (CalendarRound): The Calendar Round position of the
                date
            glyph_g (str): The Glyph G associated with the date, e.g. "G3"

        """
        if long_count is None:
            self.LongCount = LongCount(None, None, None, None, None)
        else:
            self.long_count = long_count
            if long_count.winal is not None and long_count.kin is not None:
                g = self.long_count.get_glyph_g()
                if g != glyph_g and glyph_g is not None:
                    raise ValueError("Provided Glyph G does not match the Long Count date")
                self.glyph_g = g

            else:
                self.glyph_g = glyph_g

        if calendar_round is None:
            if long_count is not None:
                if not self.long_count.has_missing():
                    self.calendar_round = self.long_count.get_calendar_round()
                else:
                    self.calendar_round = CalendarRound(None, None)
        else:
            self.calendar_round = calendar_round

    def has_missing(self):
        """ Checks whether the Mayadate object has missing values in any position

        Returns:
            (bool): True if any of the Long Count components (baktun, katun, ...)
                or Calendar Round components are None. Otherwise returns False.

        """
        return (self.long_count.has_missing() or self.calendar_round.has_missing())

    def add_days(self, num_days, in_place=False):
        """Adds num_days days (kin) to the current Mayadate object

        Args:
            num_days (int): The number of days to add
            in_place (bool): If True, modify the existing Mayadate object, else
                return a new Mayadate object. Defaults to False.

        Returns:
            (Mayadate): The Mayadate object num_days ahead of the current Mayadate
                object's date.

        """

        if in_place:
            self.long_count = self.long_count.add_days(num_days)
            self.calendar_round = self.calendar_round.add_days(num_days)

            return self
        else:
            return Mayadate(self.long_count.add_days(num_days),
                self.calendar_round.add_days(num_days))

    def infer_long_count_dates(self):
        """Finds Long Count dates that match the supplied information

        Returns:
            (list) A list of potential Long Count dates that match the supplied
                portions of the Long Count and Calendar Round Dates

        """

        if not self.long_count.has_missing():
            return [self.long_count]

        poss_lc = self.__infer_lc_recursive(self.long_count.to_list(), [])

        if poss_lc == []:
            print("No matching dates found - check the inputted values")

        return poss_lc

    def infer_mayadates(self):
        """Finds Maya calendar dates that match the supplied information

        Returns:
            (list) A list of potential Mayadate objects that match the supplied
                portions of the Long Count and Calendar Round Dates

        """
        if not self.long_count.has_missing():
            return [self.long_count.get_mayadate()]

        lcs = self.infer_long_count_dates()
        if lcs == []:
            print("No matching dates found - check the inputted values")
        return [lc.get_mayadate() for lc in lcs]

    def __infer_lc_recursive(self, lc, poss_dates):
        """ Helper function to recursively check for possible dates """

        if None not in lc:
            lc_obj = LongCount(*lc)
            if self.calendar_round.match(lc_obj.get_calendar_round()):
                if self.glyph_g is not None:
                    if lc_obj.get_glyph_g() == self.glyph_g:
                        return lc_obj
                else:
                    return lc_obj
            return

        max_vals = [13,20,20,18,20]

        for idx, v in enumerate(zip(lc, max_vals)):
            val, max = v
            if val is None:
                for i in range(max):

                    lc_test = lc[:]
                    lc_test[idx] = i

                    res = self.__infer_lc_recursive(lc_test, poss_dates)
                    if type(res) is LongCount:
                        poss_dates.append(res)
                break

        return poss_dates

    def to_julian_day(self, correlation=584283):
        """Converts the Mayan calendar date to its corresponding Julian Day number

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (int): The Julian Day number associated with the Mayan calendar date

        """
        return self.long_count.to_julian_day(correlation)

    def to_julian(self, correlation=584283):
        """Converts the Mayan calendar date to its corresponding Julian calendar date

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (JulianDate): The Julian calendar date associated with the Mayan
                calendar date

        """
        return self.long_count.to_julian(correlation)

    def to_gregorian(self, correlation=584283):
        """Converts the Mayan calendar date to its corresponding Gregorian calendar date

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (int): The Gregorian calendar date associated with the Mayan
                calendar date

        """
        return self.long_count.to_gregorian(correlation)

    def get_total_kin(self):
        """ Returns the total number of kin since the initial date 0.0.0.0.0

        Returns:
            (int): The number of kin since the initial date of the Mayan calendar.

        """
        return self.long_count.get_total_kin()

    def get_glyph_g(self):
        """Calculates the number of the Glyph G associated with the Long Count date

        Returns:
            (str): The Glyph G associated with the given date e.g. "G6"

        """

        return self.long_count.get_glyph_g()

    def __add__(self, dist):
        lc = self.long_count + dist.long_count

        return lc.get_mayadate()

    def __sub__(self, dist):
        dist = self.long_count - dist.long_count

        return dist

    def __eq__(self, date):
        if self.get_total_kin() == date.get_total_kin():
            return True
        else:
            return False

    def __gt__(self, date):
        if self.get_total_kin() > date.get_total_kin():
            return True
        else:
            return False

    def __ge__(self, date):
        if self.get_total_kin() >= date.get_total_kin():
            return True
        else:
            return False

    def __lt__(self, date):
        if self.get_total_kin() < date.get_total_kin():
            return True
        else:
            return False

    def __le__(self, date):
        if self.get_total_kin() <= date.get_total_kin():
            return True
        else:
            return False

    def __repr__(self):
        return f"{self.long_count.__repr__()}  {self.calendar_round.__repr__()}"


def main():
    pass

if __name__ == "__main__":
    main()




from .utils.mayadate import Mayadate
from .utils.long_count import LongCount, DistanceNumber, kin_to_long_count
from .utils.calendar_round import CalendarRound
from .utils.haab import Haab, HAAB_MONTHS
from .utils.tzolkin import Tzolkin, TZOLKIN_DAYS
from .utils import *

__all__ = ['Mayadate', 'LongCount', 'DistanceNumber',
    'CalendarRound', 'Haab', 'Tzolkin', 'utils', 'kin_to_long_count',
    'HAAB_MONTHS', 'TZOLKIN_DAYS']



import math
import datetime

from .calendar_round import *
from .long_count import *
from .tzolkin import *
from .haab import *
from .mayadate import *

class JulianDate:
    """Basic class to handle (proleptic) Julian calendar dates and conversions

    Note that this class uses the astronomical year convention for years before 1 CE,
    i.e. 1 BCE = 0, 2 BCE = -1, etc.

    Attributes:
        day (int): The day of the Julian calendar date
        month (int): The month number of the Julian calendar date
        year (int): The (astronomical) year number of the Julian calendar date
    """

    def __init__(self, day, month, year):
        """Creates a new JulianDate object

        Note that this class uses the astronomical year convention for years before 1 CE,
        i.e. 1 BCE = 0, 2 BCE = -1, etc.

        Args:
            day (int): The day of the Julian calendar date
            month (int): The month number of the Julian calendar date
            year (int): The (astronomical) year number of the Julian calendar date
        """

        self.day = day
        self.month = month
        self.year = year

    def to_julian_day(self):
        """Converts the Julian Calendar date to its corresponding Julian Day number

        Note that the algorithm is only valid for Julian Day numbers greater than or
        equal to zero, i.e. Julian calendar years after -4712 (4713 BCE). Earlier
        calendar years will raise a ValueError.

        Adapted from: https://www.researchgate.net/publication/316558298_Date_Algorithms#pf5

        Returns:
            (int): The Julian Day number corresponding to the Julian Calendar date

        """

        if self.year < -4712:
            raise ValueError("Algorithm only valid for Julian year greater than or equal to -4712")

        if self.month < 3:
            M = self.month + 12
            Y = self.year - 1
        else:
            M = self.month
            Y = self.year

        D = self.day

        return D + (153 * M - 457) // 5 + 365 * Y + math.floor(Y / 4) + 1721116.5

    def to_gregorian(self, as_datetime=False):
        """Converts the Julian calendar date to its Gregorian calendar equivalent

        Returns:
            (GregorianDate): The Gregorian calendar date corresponding to the
                Julian calendar date.

        """
        return julian_day_to_gregorian(round(self.to_julian_day()))


    def to_mayadate(self, correlation=584283):
        """Converts the Julian calendar date to its Mayan calendar equivalent

        Returns:
            (Mayadate): The Mayan calendar date corresponding to the Gregorian
                calendar date.

        """

        from .long_count import LongCount, kin_to_long_count
        from .mayadate import Mayadate

        num_kin = round(self.to_julian_day()) - correlation
        long_count = kin_to_long_count(num_kin)

        return Mayadate(long_count, None)

    def __repr__(self):
        return f"({self.day}, {self.month}, {self.year})"

    def __str__(self):
        if self.year > 0:
            return f"{_num_to_month(self.month)} {self.day}, {self.year} CE"

        elif self.year <= 0:
            return f"{_num_to_month(self.month)} {self.day}, {abs(self.year) + 1} BCE"


class GregorianDate:
    """Basic class to handle (proleptic) Gregorian calendar dates and conversions

    Note that this class uses the astronomical year convention for years before 1 CE,
    i.e. 1 BCE = 0, 2 BCE = -1, etc.

    Attributes:
        day (int): The day of the Gregorian calendar date
        month (int): The month number of the Gregorian calendar date
        year (int): The (astronomical) year number of the Gregorian calendar date
    """

    def __init__(self, day, month, year):
        """Creates a new GregorianDate object

        Note that this class uses the astronomical year convention for years before 1 CE,
        i.e. 1 BCE = 0, 2 BCE = -1, etc.

        Args:
            day (int): The day of the Gregorian calendar date
            month (int): The month number of the Gregorian calendar date
            year (int): The (astronomical) year number of the Gregorian calendar date
        """

        self.day = day
        self.month = month
        self.year = year

    def to_julian_day(self):
        """Converts the Gregorian calendar date to its Julian Day number equivalent

        Adapted from: https://www.researchgate.net/publication/316558298_Date_Algorithms#pf5

        Returns:
            (float): The Julian day number corresponding to the Gregorian calendar
                date.

        """
        if self.month < 3:
            M = self.month + 12
            Y = self.year - 1
        else:
            M = self.month
            Y = self.year

        D = self.day

        return D + (153 * M - 457) // 5 + 365 * Y + math.floor(
            Y / 4) - math.floor(Y / 100) + math.floor(Y / 400) + 1721118.5

    def to_julian(self):
        """Converts the Gregorian calendar date to its Julian calendar equivalent

        Returns:
            (JulianDate): The Julian calendar date corresponding to the Gregorian
                calendar date.

        """
        return julian_day_to_julian(round(self.to_julian_day))

    def to_mayadate(self, correlation=584283):
        """Converts the Gregorian calendar date to its Mayan calendar equivalent

        Returns:
            (Mayadate): The Mayan calendar date corresponding to the Gregorian
                calendar date.

        """
        from .long_count import LongCount, kin_to_long_count
        from .mayadate import Mayadate

        num_kin = round(self.to_julian_day()) - correlation
        long_count = kin_to_long_count(num_kin)

        return Mayadate(long_count, None)

    def to_datetime(self):
        """Converts the GregorianDate object to a datetime.date object

        Note that datetime.date objects do not support years before 1 CE. Attempting
        to convert GregorianDate objects with year before 1 CE will raise a ValueError.

        Returns:
            (datetime.date): The datetime.date object corresponding to the Gregorian
                calendar date.

        """
        if self.year < 1:
            raise ValueError("datetime.date objects do not support years before 1 CE")

        return datetime.date(self.year, self.month, self.day)

    def __repr__(self):
        return f"({self.day}, {self.month}, {self.year})"

    def __str__(self):
        if self.year > 0:
            return f"{_num_to_month(self.month)} {self.day}, {self.year} CE"

        elif self.year <= 0:
            return f"{_num_to_month(self.month)} {self.day}, {abs(self.year) + 1} BCE"




def _convert_julian_day(julian_day, mode='julian'):
    """Converts a Julian Day number to its (proleptic) Julian or Gregorian calendar equivalent

    Adapted from: https://en.wikipedia.org/wiki/Julian_day#Julian_or_Gregorian_calendar_from_Julian_day_number

    Note that the algorithm is only valid for Julian Day numbers greater than or
    equal to zero. Negative arguments for julian_day will raise a ValueError.

    Args:
        julian_day (int): Julian Day number to convert, must be greater than or
            equal to 0
        mode (str): The target calendar to convert to, either 'julian' or
            'gregorian'. Defaults to 'julian'.

    Returns:
        A (day, month, year) tuple representing the day, month, and year in the
            target calendar.
    """
    if julian_day < 0:
        raise ValueError("Algorithm only valid for Julian Day greater than or equal to zero")
    julian_day = round(julian_day)

    # algorithm parameters
    y = 4716
    j = 1401
    m = 2
    n = 12
    r = 4
    p = 1461
    v = 3
    u = 5
    s = 153
    w = 2
    B = 274277
    C = -38

    # intermediate calculations
    if mode == 'julian':
        f = julian_day + j
    elif mode == 'gregorian':
        f = julian_day + j + (((4 * julian_day + B) // 146097) * 3) // 4 + C
    else:
        raise ValueError("Unrecognized mode - supports 'julian' or 'gregorian'")

    e = r * f + v
    g = (e % p) // r
    h = u * g + w

    day = (h % s) // u + 1  # day in target calendar
    month =  ((h // s + m) % n) + 1 # month in target calendar
    year = (e // p) - (y + (n + m - month) // n) # year in target calendar

    return day, month, year



def julian_day_to_julian(julian_day):
    """Converts a Julian Day number to its (proleptic) Julian calendar equivalent

    Adapted from: https://en.wikipedia.org/wiki/Julian_day#Julian_or_Gregorian_calendar_from_Julian_day_number

    Note that the algorithm is only valid for Julian Day numbers greater than or
    equal to zero. Negative arguments for julian_day will raise a ValueError.

    Args:
        julian_day (int): Julian Day number to convert, must be greater than or
            equal to 0

    Returns:
        A (day, month, year) tuple representing the day, month, and year in the
            Julian calendar.

    """

    day, month, year = _convert_julian_day(julian_day, mode='julian')

    return JulianDate(day, month, year)



def julian_day_to_gregorian(julian_day):
    """Converts a Julian Day number to its (proleptic) Gregorian calendar equivalent

    Adapted from: https://en.wikipedia.org/wiki/Julian_day#Julian_or_Gregorian_calendar_from_Julian_day_number

    Note that the algorithm is only valid for Julian Day numbers greater than or
    equal to zero. Negative arguments for julian_day will raise a ValueError.

    Args:
        julian_day (int): Julian Day number to convert, must be greater than or
            equal to 0

    Returns:
        A (day, month, year) tuple representing the day, month, and year in the
            Gregorian calendar.

    """
    day, month, year = _convert_julian_day(julian_day, mode='gregorian')

    return GregorianDate(day, month, year)

def datetime_to_gregorian(date):
    """Converts a datetime.date object to a GregorianDate object

    Args:
        date (datetime.date): The datetime.date object to convert

    Returns:
        (GregorianDate): The corresponding GregorianDate object

    """

    return GregorianDate(date.day, date.month, date.year)

def datetime_to_julian(date):
    """Converts a datetime.date object to the corresponding Julian calendar date

    Args:
        date (datetime.date): The datetime.date object to convert

    Returns:
        (JulianDate): The corresponding Julian calendar date

    """

    g = GregorianDate(date.day, date.month, date.year)
    return g.to_julian()

def datetime_to_julian_day(date):
    """Converts a datetime.date object to the corresponding Julian Day number

    Args:
        date (datetime.date): The datetime.date object to convert

    Returns:
        (float): The corresponding Julian Day number

    """

    g = GregorianDate(date.day, date.month, date.year)
    return g.to_julian_day()

def datetime_to_mayadate(date):
    """Converts a datetime.date object to the corresponding Maya calendar date

    Args:
        date (datetime.date): The datetime.date object to convert

    Returns:
        (Mayadate): The corresponding Mayan calendar date.

    """

    g = GregorianDate(date.day, date.month, date.year)
    return g.to_mayadate()



def _num_to_month(num):
    """Helper function to convert month number to short name

    Args:
        num (int): the month number to convert

    Returns:
        (str): The three letter short name of the corresponding month

    """

    return {1 : 'Jan',
            2 : 'Feb',
            3 : 'Mar',
            4 : 'Apr',
            5 : 'May',
            6 : 'Jun',
            7 : 'Jul',
            8 : 'Aug',
            9 : 'Sep',
            10 : 'Oct',
            11 : 'Nov',
            12 : 'Dec'}[num]




__all__ = ['Tzolkin', 'TZOLKIN_DAYS', 'TZOLKIN_DAY_TO_NUM', 'TZOLKIN_IDX_TO_DAY',
'TZOLKIN_NUM_TO_DAY', 'TZOLKIN_DAY_TO_IDX']
# Module level constants
TZOLKIN_DAYS = ["Imix", "Ik", "Akbal", "Kan", "Chikchan", "Kimi", "Manik",
                    "Lamat", "Muluk", "Ok", "Chuwen", "Eb", "Ben", "Ix",
                    "Men", "Kib", "Kaban", "Etznab", "Kawak", "Ajaw"]

TZOLKIN_IDX_TO_DAY = {idx:day for idx, day in enumerate(TZOLKIN_DAYS)}
TZOLKIN_DAY_TO_IDX = {day:idx for idx, day in TZOLKIN_IDX_TO_DAY.items()}

TZOLKIN_NUM_TO_DAY = {}

for _i in range(260):
    _date = ((_i % 13) + 1, TZOLKIN_IDX_TO_DAY[(_i % 20)])
    TZOLKIN_NUM_TO_DAY[_i] = _date

TZOLKIN_DAY_TO_NUM = {date:num for num, date in TZOLKIN_NUM_TO_DAY.items()}


class Tzolkin:
    """Represents a day number, day name combination in the 260 day count

    Attributes:
        day_number (int): The day number associated with the Tzolkin date
        day_name (str): The day name associated with the Tzolkin date
        tzolkin_num (int): Integer from 0-259 representing days since 1 Imix.

    """

    def __init__(self, day_number=None, day_name=None, tzolkin_num=None):
        """Creates a new Tzolkin object

        Can either be constructed from a day name, day number combination or
        from the position in the tzolkin count counting from 1 Imix.

        Args:
            day_number (int): Integer from 1-13 representing the day number
            day_name (str): Day name


        """

        if tzolkin_num is not None:
            # check if day name/number matches tzolkin number if provided
            if day_number is not None and day_name is not None:
                implied_num = TZOLKIN_DAY_TO_NUM[(day_number, day_name)]
                if implied_num != tzolkin_num:
                    raise ValueError(f"Provided Tzolkin number {tzolkin_num} does not match provided day name and number {day_number} {day_name}")

            self.reset_by_tzolkin_num(tzolkin_num)


        else:
            if day_name not in TZOLKIN_DAYS and day_name is not None:
                raise ValueError(f"Invalid tzolkin day name {day_name}")
            self.day_name = day_name

            if day_number not in list(range(1, 13)) and day_number is not None:
                raise ValueError("Invalid tzolkin day number - must be integer between 1 and 13")
            self.day_number = day_number

            if day_number is not None and day_name is not None:
                self.tzolkin_num = TZOLKIN_DAY_TO_NUM[(day_number, day_name)]

    def reset_by_tzolkin_num(self, new_num):
        """ Set the Tzolkin object to a new position by its 260 day count number

        Note:
            1 Imix is used as the reference 'Day 0' of the cycle

        Args:
            new_num (int): Integer from 0-259 representing new position in the
                260 day count.
        """

        if new_num >= 260:
            raise ValueError("Invalid Tzolkin number, must be between 0 and 259")

        self.tzolkin_num = new_num
        self.day_number, self.day_name = TZOLKIN_NUM_TO_DAY[self.tzolkin_num]

    def add_days(self, num_days, in_place=False):
        """ Adds days to the current Tzolkin object

        Args:
            num_days (int): Number of days to add to the Tzolkin object
            in_place (bool): Whether to modify the existing object or return a
                new object. Defaults to False.

        Returns:
            A new Tzolkin object num_days ahead of the previous object

        """
        new_num = (self.tzolkin_num + num_days) % 260
        if in_place:
            self.reset_by_tzolkin_num(new_num)
            return self

        else:
            return Tzolkin(tzolkin_num=new_num)

    def has_missing(self):
        """ Checks whether the day number or name is missing

        Returns:
            (bool): True if either the day number or day name is None, False
                otherwise

        """
        if self.day_name is None or self.day_number is None:
            return True

        return False

    def match(self, date):
        """Checks for a potential match with another Tzolkin object

        A value of None is treated as matching any value, consistent with the use
        of None to mark values for later inference.

        Args:
            date (Tzolkin): The Tzolkin object to check for a match with

        Returns:
            (bool): True if the day name and number match, with None as an
                automatic match. False otherwise.

        """
        name_same = self.__fuzzy_eq(self.day_name, date.day_name)
        num_same = self.__fuzzy_eq(self.day_number, date.day_number)

        if name_same and num_same:
            return True

        else:
            return False

    def __fuzzy_eq(self, v1, v2):
        """ Helper function for NoneType matching """

        if v1 == v2 or v1 is None or v2 is None:
            return True

        return False


    def __eq__(self, date):
        name_same = (self.day_name == date.day_name)
        num_same = (self.day_number == date.day_number)

        if name_same and num_same:
            return True

        else:
            return False

    def __sub__(self, date):
        return abs(self.tzolkin_num - date.tzolkin_num)

    def __repr__(self):
        return f"{self.day_number} {self.day_name}"



from .utils import *
from .calendar_round import *
from .long_count import *
from .tzolkin import *
from .haab import *
from .mayadate import *

__all__ = ['LongCount', 'DistanceNumber', 'kin_to_long_count']

class LongCount:
    """Represents a position in the Maya Long Count

    Valid for dates from 0.0.0.0.0 to 19.19.19.17.19. Use NoneType to mark
    missing positions in the date for later inference.

    Attributes:
        baktun (int or NoneType): The Bak'tun number of the Long Count date.
            Integer between 0 and 19 or None.
        katun (int or NoneType): The K'atun number of the Long Count date.
            Integer between 0 and 19 or None.
        tun (int or NoneType): The Tun number of the Long Count date.
            Integer between 0 and 19 or None.
        winal (int or NoneType): The Winal number of the Long Count date.
            Integer between 0 and 17 or None.
        kin (int or NoneType): The Kin number of the Long Count date.
            Integer between 0 and 19 or None.

    """

    def __init__(self, baktun=0, katun=0, tun=0, winal=0, kin=0):
        """Creates a new LongCount object

        Use NoneType to mark missing positions in the date for later inference.

        Args:
            baktun (int or NoneType): The Bak'tun number of the Long Count date.
                Integer between 0 and 19 or None.
            katun (int or NoneType): The K'atun number of the Long Count date.
                Integer between 0 and 19 or None.
            tun (int or NoneType): The Tun number of the Long Count date.
                Integer between 0 and 19 or None.
            winal (int or NoneType): The Winal number of the Long Count date.
                Integer between 0 and 17 or None.
            kin (int or NoneType): The Kin number of the Long Count date.
                Integer between 0 and 19 or None.

        """
        if baktun is not None and baktun >= 20:
            raise ValueError("Baktun must be between 0 and 19 or NoneType")
        self.baktun = baktun
        if katun is not None and katun >= 20:
            raise ValueError("Katun must be between 0 and 19 or NoneType")
        self.katun = katun

        if tun is not None and tun >= 20:
            raise ValueError("Tun must be between 0 and 19")
        self.tun = tun

        if winal is not None and winal >= 18:
            raise ValueError("Winal must be between 0 and 17")
        self.winal = winal

        if kin is not None and kin >= 20:
            raise ValueError("Kin must be between 0 and 19")
        self.kin = kin


    def get_total_kin(self):
        """ Returns the total number of kin since the initial date 0.0.0.0.0

        Returns:
            (int): The number of kin since the initial date of the Mayan calendar.

        """

        if self.has_missing():
            raise ValueError("Operation not valid for incomplete Long Count dates, try inferring the missing portions")

        else:
            total_kin = (self.kin + (self.winal * 20) + (self.tun * 20 * 18) +
                (self.katun * 18 * (20 ** 2)) + (self.baktun * 18 * (20 ** 3)))

            return total_kin


    def get_calendar_round(self):
        """Returns the calendar round associated with the current LongCount object

        Returns:
            (CalendarRound): The position in the calendar round corresponding to
                the Long Count date

        """

        # Mythological start date
        initial_date = CalendarRound(Tzolkin(4, "Ajaw"), Haab(8, "Kumku"))
        initial_date.add_days(self.get_total_kin(), True)

        return initial_date

    def get_mayadate(self):
        """Returns a Mayadate object from the current LongCount object

        Returns:
            (Mayadate): The Long Count with its associated Calendar Round position

        """
        from .mayadate import Mayadate

        if self.has_missing():
            return Mayadate(self, None)

        return Mayadate(self, self.get_calendar_round())


    def add_days(self, num_days, in_place=False):
        """Adds num_days days (kin) to the current LongCount object

        Args:
            num_days (int): The number of days to add
            in_place (bool): If True, modify the existing LongCount object, else
                return a new LongCount object. Defaults to False.

        Returns:
            (LongCount): The LongCount object num_days ahead of the current LongCount
                object's date.

        """
        total_days = self.get_total_kin() + num_days
        if in_place:
            self = kin_to_long_count(total_days)
            return self
        else:
            return kin_to_long_count(total_days)


    def get_glyph_g(self):
        """Calculates the number of the Glyph G associated with the Long Count date

        Returns:
            (str): The Glyph G associated with the given date e.g. "G6"

        """

        if self.winal is None or self.kin is None:
            raise ValueError("Both the Winal and Kin numbers must be provided to infer Glyph G")

        g_num = (self.winal * 20 + self.kin) % 9

        if g_num == 0:
            g_num = 9

        return f"G{g_num}"

    def to_list(self):
        """Returns a list representation of the LongCount object

        Returns:
            (list): A list of integers in the form [baktun, katun, tun, winal, kin]

        """

        return [self.baktun, self.katun, self.tun, self.winal, self.kin]

    def to_julian_day(self, correlation=584283):
        """Converts the Long Count date to its corresponding Julian Day number

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (int): The Julian Day number associated with the Long Count date

        """

        return self.get_total_kin() + correlation

    def to_julian(self, correlation=584283):
        """Converts the Long Count date to its corresponding Julian calendar date

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (JulianDate): The Julian calendar date associated with the Long Count date

        """

        julian_day = self.to_julian_day(correlation)

        return julian_day_to_julian(julian_day)

    def to_gregorian(self, correlation=584283):
        """Converts the Long Count date to its corresponding Gregorian calendar date

        By default uses the correlation constant 584,283 proposed by Thompson.

        Args:
            correlation (int): The correlation constant to use in the conversion.
                Defaults to 584283.

        Returns:
            (GregorianDate): The Gregorian calendar date associated with the Long
                Count date

        """

        julian_day = self.to_julian_day(correlation)

        return julian_day_to_gregorian(julian_day)

    def has_missing(self):
        """ Checks whether the Long Count object has missing values in any position

        Returns:
            (bool): True if any of the Long Count components (baktun, katun, ...)
                are None. Otherwise returns False.

        """

        if self.baktun is None:
            return True

        if self.katun is None:
            return True

        if self.tun is None:
            return True

        if self.winal is None:
            return True

        if self.kin is None:
            return True

        return False


    def __add__(self, dist):

        kin, carry = self.__add_and_carry(self.kin, dist.kin, 0, 20)
        winal, carry = self.__add_and_carry(self.winal, dist.winal, carry, 18)
        tun, carry = self.__add_and_carry(self.tun, dist.tun, carry, 20)
        katun, carry = self.__add_and_carry(self.katun, dist.katun, carry, 20)
        baktun, carry = self.__add_and_carry(self.baktun, dist.baktun, carry, 20)

        return LongCount(baktun, katun, tun, winal, kin)

    def __add_and_carry(self, val_1, val_2, carry, max):
        raw_sum = val_1 + val_2 + carry

        return raw_sum % max, int(raw_sum >= max)


    def __sub__(self, dist):
        kin_diff = self.get_total_kin() - dist.get_total_kin()
        if kin_diff >= 0:
            return DistanceNumber(kin_to_long_count(kin_diff), sign=1)
        else:
            return DistanceNumber(kin_to_long_count(kin_diff * -1), sign=-1)

    def __eq__(self, date):
        if self.get_total_kin() == date.get_total_kin():
            return True
        else:
            return False

    def __gt__(self, date):
        if self.get_total_kin() > date.get_total_kin():
            return True
        else:
            return False

    def __ge__(self, date):
        if self.get_total_kin() >= date.get_total_kin():
            return True
        else:
            return False

    def __lt__(self, date):
        if self.get_total_kin() < date.get_total_kin():
            return True
        else:
            return False

    def __le__(self, date):
        if self.get_total_kin() <= date.get_total_kin():
            return True
        else:
            return False

    def __repr__(self):
        return f"{self.baktun}.{self.katun}.{self.tun}.{self.winal}.{self.kin}"

class DistanceNumber(LongCount):
    """ Represents a signed Distance Number in Long Count units


    Attributes:
        long_count (LongCount): The Distance Number in Long Count units
        sign (int): 1 for a positive number, -1 for a negative number

    """

    def __init__(self, long_count=LongCount(), sign=1):
        """Creates a new DistanceNumber object


        Args:
            long_count (LongCount): The Distance Number in LongCount units
            sign (int): 1 for a positive number, -1 for a negative number

        """
        self.long_count = long_count
        super().__init__(long_count.baktun, long_count.katun, long_count.tun,
            long_count.winal, long_count.kin)
        if sign == 1 or sign == -1:
            self.sign = sign
        else:
            raise ValueError("sign must be either 1 or -1")

    def get_total_kin(self):
        """ Returns the total number of kin associated with the distance number

        Returns:
            (int): The number of kin associated with the distance number.

        """
        return  self.sign * super().get_total_kin()

    def to_approx_years(self, pretty_print=False):
        total_kin = abs(self.get_total_kin())

        years = int(total_kin // 365.25)
        total_kin -= years * 365.25

        months = int(total_kin  // 30.44) #average month length
        total_kin -= months * 30.44

        if not pretty_print:
            return self.sign * years, self.sign * months, self.sign * round(total_kin)

        if self.sign == 1:
            return f"{years} years, {months} months, {round(total_kin)} days"
        else:
            return f"-{years} years, -{months} months, -{round(total_kin)} days"



    def __repr__(self):
        if self.sign == 1:
            return self.long_count.__repr__()

        else:
            return f"-{self.long_count.__repr__()}"



def kin_to_long_count(num_kin):
    """Converts the given number of days (kin) to a Long Count date

    Assumes the number of kin is counted from the Maya zero date of 0.0.0.0.0
    4 Ajaw 8 Kumk'u. For adding days to another date, use the add_days method
    of the LongCount object.

    Args:
        num_kin (int): The number of kin since the Maya zero date 0.0.0.0.0

    Returns:
        (LongCount): A Long Count date representing the position indicated by
            the given number of days after the zero date


    """

    if type(num_kin) is not int:
        num_kin = int(num_kin)

    long_count = LongCount()

    long_count.baktun = num_kin // (18 * (20 ** 3))
    num_kin = num_kin - (long_count.baktun * 18 * (20 ** 3))

    long_count.katun = num_kin // (18 * (20 ** 2))
    num_kin = num_kin - (long_count.katun * 18 * (20 ** 2))

    long_count.tun = num_kin // (18 * 20)
    num_kin = num_kin - (long_count.tun * 18 * 20)

    long_count.winal = num_kin // 20
    num_kin = num_kin - (long_count.winal * 20)

    long_count.kin = num_kin

    return long_count





__all__ = ['Haab', 'HAAB_MONTHS', 'HAAB_IDX_TO_MONTH', 'HAAB_MONTH_TO_IDX']
# Module level constants
HAAB_MONTHS = ["Pop", "Wo", "Sip", "Sotz", "Sek", "Xul", "Yaxkin", "Mol", "Chen",
                "Yax", "Sak", "Keh", "Mak", "Kankin", "Muwan", "Pax", "Kayab",
                "Kumku", "Wayeb"]

HAAB_IDX_TO_MONTH = {idx:month for idx, month in enumerate(HAAB_MONTHS)}
HAAB_MONTH_TO_IDX = {month:idx for idx, month in HAAB_IDX_TO_MONTH.items()}


class Haab:
    """Represents a month number, month name combination in the 365 day count

    Attributes:
        day_number (int): The day number associated with the Haab date
        day_name (str): The day name associated with the Haab date

    """
    def __init__(self, month_number=None, month_name=None):
        """Creates a new Haab object

        Can either be constructed from a day name, day number combination or
        from the position in the tzolkin count counting from 1 Imix.

        Args:
            month_number (int): Integer from 0-19 (or 0-4 for Wayeb)
                representing the month number
            month_name (str): month name


        """

        if month_name not in HAAB_MONTHS and month_name is not None:
            raise ValueError(f"Invalid Haab month name {month_name}")
        self.month_name = month_name

        if month_name == "Wayeb":
             if month_number not in list(range(5)):
                 raise ValueError("Invalid Haab month number, Wayeb number must be between 0 and 4")
        elif month_number not in list(range(20)) and month_number is not None:
            raise ValueError("Invalid Haab month number, must be an integer between 0 and 19 or NoneType")

        self.month_number = month_number

        if not self.has_missing():
            self.haab_num = 20 * HAAB_MONTH_TO_IDX[month_name] + month_number
        else:
            self.haab_num = None

    def has_missing(self):
        """ Checks whether the month number or name is missing

        Returns:
            (bool): True if either the month number or month name is None, False
                otherwise

        """
        if self.month_name is None or self.month_number is None:
            return True

        return False

    def reset_by_haab_num(self, new_num):
        """ Set the Haab object to a new position by its 365 day count number

        Note:
            0 Pop is used as the reference 'Day 0' of the cycle

        Args:
            new_num (int): Integer from 0-365 representing new position in the
                365 day count.
        """
        self.month_name = HAAB_IDX_TO_MONTH[new_num // 20]
        self.month_number = new_num % 20

        return self

    def add_days(self, num_days, in_place=False):
        """ Adds days to the current Haab object

        Args:
            num_days (int): Number of days to add to the Haab object
            in_place (bool): Whether to modify the existing object or return a
                new object. Defaults to False.

        Returns:
            A new Haab object num_days ahead of the previous object

        """
        new_num = (self.haab_num + num_days) % 365
        if in_place:
            self.haab_num = new_num
            self.reset_by_haab_num(self.haab_num)

            return self
        else:
            return Haab().reset_by_haab_num(new_num)

    def match(self, date):
        """Checks for a potential match with another Haab object

        A value of None is treated as matching any value, consistent with the use
        of None to mark values for later inference.

        Args:
            date (Haab): The Haab object to check for a match with

        Returns:
            (bool): True if the month name and number match, with None as an
                automatic match. False otherwise.

        """

        name_same = self.__fuzzy_eq(self.month_name, date.month_name)
        num_same = self.__fuzzy_eq(self.month_number, date.month_number)

        if name_same and num_same:
            return True
        else:
            return False

    def __fuzzy_eq(self, v1, v2):
        """ Helper function for NoneType matching """

        if v1 == v2 or v1 is None or v2 is None:
            return True

        return False

    def __eq__(self, date):
        name_same = self.month_name == date.month_name
        num_same = self.month_number == date.month_number

        if name_same and num_same:
            return True
        else:
            return False

    def __sub__(self, date):
        return abs(self.haab_num - date.haab_num)


    def __repr__(self):
        return f"{self.month_number} {self.month_name}"



from .haab import *
from .tzolkin import *

__all__ =['CalendarRound']

class CalendarRound:
    """Represents a position in the Mayan Calendar Round.

    Supports use of NoneType to mark positions for later inference.

    Attributes:
        tzolkin (Tzolkin): The tzolkin date portion of the the Calendar Round position
        haab (Haab): The haab date portion of the the Calendar Round position
        valid (bool): Whether the Tzolkin day name matches the Haab month number

    """
    def __init__(self, tzolkin=None, haab=None, override_coef_check=False):
        """ Creates a new CalendarRound object

        Supports use of NoneType to mark positions for later inference.

        Note that each Tzolkin day name can only occur with a specific set of Haab
        month coefficients. Throws a ValueError if these rules are violated. Can
        override these rules (for example for work with Post Classic dates) by
        setting override_coef_check to True.

        Args:
            tzolkin (Tzolkin or NoneType): The tzolkin date portion of the the
                Calendar Round position. Use None to denote missing Tzolkin date.
            haab (Haab or NoneType): The haab date portion of the the Calendar
                Round position. Use None to denote missing Haab date.
            override_coef_check (bool): Whether to override the check for valid
                month number, day name combinations. Defaults to False.

        """

        if tzolkin is None:
            self.tzolkin = Tzolkin(None, None)
        else:
            self.tzolkin = tzolkin

        if haab is None:
            self.haab = Haab(None, None)
        else:
            self.haab = haab

        self.valid = self.__check_valid()

        if not override_coef_check and not self.valid:
            raise ValueError("Invalid Haab month coefficient, Tzolkin day name combo")


    def has_missing(self):
        """Checks whether the Calendar Round has any missing components

        Returns:
            (bool): True if the Tzolkin or Haab contain missing components.
                False otherwise.

        """

        return (self.tzolkin.has_missing() or self.haab.has_missing())


    def add_days(self, num_days, in_place=False):
        """Adds num_days days (kin) to the current CalendarRound object

        Args:
            num_days (int): The number of days to add
            in_place (bool): If True, modify the existing CalendarRound object,
                else return a new CalendarRound object. Defaults to False.

        Returns:
            (CalendarRound): The CalendarRound object num_days ahead of the current
                CalendarRound object's date.

        """
        if in_place:
            self.haab.add_days(num_days, True)
            self.tzolkin.add_days(num_days, True)

            return self

        else:
            new_haab = self.haab.add_days(num_days)
            new_tzolkin = self.tzolkin.add_days(num_days)

            return CalendarRound(new_tzolkin, new_haab)


    def __check_valid(self):
        """Checks whether the Tzolkin day name can occur with the Haab month number

        Returns:
            (bool): True if the Tzolkin day name can occur with the Haab month
                number. False otherwise.

        """

        if self.haab.month_number is None or self.tzolkin.day_name is None:
            return True

        if self.haab.month_number % 5 == 0:
            if self.tzolkin.day_name not in ("Kaban", "Ik", "Manik", "Eb"):
                return False

            return True

        elif self.haab.month_number % 5 == 1:
            if self.tzolkin.day_name not in ("Etznab", "Akbal", "Lamat", "Ben"):
                return False
            return True

        elif self.haab.month_number % 5 == 2:
            if self.tzolkin.day_name not in ("Kawak", "Kan", "Muluk", "Ix"):
                return False
            return True

        elif self.haab.month_number % 5 == 3:
            if self.tzolkin.day_name not in ("Ajaw", "Chikchan", "Ok", "Men"):
                return False
            return True

        elif self.haab.month_number % 5 == 4:
            if self.tzolkin.day_name not in ("Imix", "Kimi", "Chuwen", "Kib"):
                return False
            return True


        else:
            raise ValueError(f"Invalid month coefficient {self.haab.month_number}")


    def get_long_count_possibilities(self, min_date, max_date):
        """Finds Long Count dates that correspond to the Calendar Round date

        Finds all Long Count dates that correspond to the Calendar Round date
        between min_date and max_date. Note that the Calendar Round cycle repeats
        every 18,980 days, i.e. the LCM of the 260 day Tzolkin cycle and the 365
        day Haab cycle.

        Args:
            min_date (LongCount): The earliest Long Count date to check
            max_date (LongCount): The latest Long Count date to check

        Returns:
            (list): A list of LongCount objects representing the possible dates

        """
        from .long_count import LongCount, kin_to_long_count

        min_cal_round = min_date.get_calendar_round()
        init_num = TZOLKIN_DAY_TO_NUM[(min_cal_round.tzolkin.day_number, min_cal_round.tzolkin.day_name)]
        day_num = TZOLKIN_DAY_TO_NUM[(self.tzolkin.day_number, self.tzolkin.day_name)]

        cycle_num = (day_num - init_num) % 260

        lc = min_date.add_days(cycle_num)

        count = 0
        while True:
            if lc.get_calendar_round().haab == self.haab:
                break
            count += 1
            lc = lc.add_days(260)


        poss_dates = []

        while True:
            if lc > max_date:
                return poss_dates

            poss_dates.append(lc)
            lc = lc.add_days(18980) #LCM of 260 and 365


    def match(self, date):
        """Checks for a potential match with another CalendarRound object

        A value of None is treated as matching any value, consistent with the use
        of None to mark values for later inference.

        Args:
            date (CalendarRound): The CalendarRound object to check for a match with

        Returns:
            (bool): True if the Tzolkin and Haab match, with None as an
                automatic match. False otherwise.

        """

        if self.tzolkin.match(date.tzolkin) and self.haab.match(date.haab):
            return True

        else:
            return False

    def __eq__(self, date):
        if self.tzolkin == date.tzolkin and self.haab == date.haab:
            return True
        else:
            return False


    def __repr__(self):
        return f"{self.tzolkin.__repr__()} {self.haab.__repr__()}"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'mayacal',
  packages = ["mayacal", "mayacal.utils"],
  version = '0.1.4-alpha',
  license='MIT',
  description = 'Basic calendar functions for the classical Mayan calendar',
  author = 'Jon Bleiberg',
  author_email="",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url = 'https://github.com/jonbleiberg88/mayacal',
  keywords = ['Maya', 'Mayan', 'Calendar', 'Classical', 'Ancient'],
  install_requires=[],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
  ],
  python_requires='>=3.5',
)
